package ucl.salinization.displays;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import repast.simphony.context.Context;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.IAction;
import repast.simphony.engine.schedule.NonModelAction;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.parameter.Parameters;
import repast.simphony.space.graph.Network;
import repast.simphony.space.projection.Projection;
import repast.simphony.visualization.DefaultDisplayData;
import repast.simphony.visualization.IDisplay;
import repast.simphony.visualization.Layout;
import repast.simphony.visualization.NullLayout;
import repast.simphony.visualization.VisualizationProperties;
import repast.simphony.visualization.gis3D.DisplayGIS3D;
import repast.simphony.visualization.gis3D.style.StyleGIS3D;
import repast.simphony.visualization.visualization2D.Display2D;
import repast.simphony.visualization.visualization2D.style.EdgeStyle2D;
import repast.simphony.visualization.visualization2D.style.Style2D;

/**
 * Custom Display Creator class for making 2D MADCABS displays.
 * 
 * @author
 *
 */
public class DisplayCreator {

	public enum DisplayType {
		STEP("Steps"), GIS("GIS");

	/**The name of the display*/
    private String name;

    /**The display constructor
     * 
     * @param name the name of the display
     */
    DisplayType(String name) {
      this.name = name;
    }

    public String getName() {
      return name;
    }
  }
	
	@NonModelAction
	/**
	 * Updates the display.
	 */
  static class DisplayUpdater implements IAction {

	/**The IDisplay object for display*/
    IDisplay display;

    /**
     * The display updater object
     * @param display a IDisplay
     */
    public DisplayUpdater(IDisplay display) {
      this.display = display;
    }
    
    /**
     * Method to execute the display.
     */
    public void execute() {
      display.update();
    }
  }


	/**
	 * Create display and register with the GUI Registry.
	 * 
	 * @param context the context in which to create the display.
	 * @param type the display type
	 */
	public IDisplay createDisplay(Context context, DisplayType type){
		
		IDisplay display = null;
		try {
			//if it is a step type then create 2D display
			if (type == DisplayType.STEP)
			  display = createDisplay2D(context);
			//otherwise it should be 3D
			else
				display = createDisplayGIS3d(context);
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		
		display.init();
    
//    // Add the display to the GUI registry.
//    RunState.getInstance().getGUIRegistry().addDisplay(type.getName(), 
//    		GUIRegistryType.DISPLAY, display);
    
//    JFrame frame = new JFrame("Map");
//    frame.setSize(800, 600);
//    frame.add(display.getPanel());
//    frame.setVisible(true);
    
    // Create the updater that will update the display, and schedule it.
    DisplayUpdater updater = new DisplayUpdater(display);   
    Parameters p = RunEnvironment.getInstance().getParameters();
//    int interval = (Integer)p.getValue("tickInterval");
    
    //schedule the update based on model time
    ScheduleParameters params = ScheduleParameters.createRepeating(0,1);
    RunState.getInstance().getScheduleRegistry().getModelSchedule().schedule(params,
            updater);
		
    return display;
	}
	
	/**
	 * Creates a display for the steps display using a custom 2D display.
	 * 
	 * @param context
	 * @return
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws ClassNotFoundException
	 */
	private Display2D createDisplay2D(Context context) throws IllegalAccessException,
		InstantiationException, ClassNotFoundException {
				
		DefaultDisplayData data = new DefaultDisplayData(context);
		data.addProjection("Steps");
		Layout layout = createLayout(context);
		
		//menu display is shown here
		PopupMenuDisplay display = new PopupMenuDisplay(data, layout);
				
		registerStyles(display);
		registerNetworkStyles(display, context);
		
		return display;
	}
	
	/**
	 * Creates a display for the GIS display.
	 * 
	 * @param context
	 * @return
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 */
	private DisplayGIS3D createDisplayGIS3d(Context context) throws ClassNotFoundException,
		IllegalAccessException, InstantiationException {
    
		DefaultDisplayData data = new DefaultDisplayData(context);
		data.addProjection("Geography");
		Layout layout = null;
  	
 		layout = new NullLayout();

 		//3D display is created here
 		DisplayGIS3D display = new DisplayGIS3D(data, layout);
  	
 		registerStyles(display);
//  	registerNetworkStyles(display);

 		return display;
  }
	

	/**
	 * Register 2D display styles
	 * 
	 * @param display
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws ClassNotFoundException
	 */
	private void registerStyles(Display2D display)throws IllegalAccessException, 
	InstantiationException, ClassNotFoundException {

		Map<String,String> classStyleMap = new LinkedHashMap<String,String>();

		// Styles need to be registered in the order in which they
		// are to be displayed, with 0 being the lowest (background).

		classStyleMap.put("anl.msc.agents.Step", "anl.msc.displays.ScenarioStyle2D");

		// Iterate through layers in order and register
		for (String agentName : classStyleMap.keySet()) {
			String styleName = classStyleMap.get(agentName);

			System.out.println(" ## registering " + agentName + " : " + styleName);
			
			try{
				Class agentClass = Class.forName(agentName, true, this.getClass().getClassLoader());
				Class styleClass = Class.forName(styleName, true, this.getClass().getClassLoader());

				Style2D style = (Style2D) styleClass.newInstance();
				display.registerStyle(agentClass, style);
			}

			catch (ClassNotFoundException e){
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Register 3DGIS display styles
	 * 
	 * @param display
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws ClassNotFoundException
	 */	
	private void registerStyles(DisplayGIS3D display) throws IllegalAccessException, 
	InstantiationException, ClassNotFoundException {

		Map<String,String> classStyleMap = new LinkedHashMap<String,String>();

		// Styles need to be registered in the order in which they
		// are to be displayed, with 0 being the lowest (background).

		classStyleMap.put("anl.msc.displays.State", "anl.msc.displays.StateStyle");
		classStyleMap.put("anl.msc.displays.City", "anl.msc.displays.CityStyle");
		
		for (String agentName : classStyleMap.keySet()) {
			String styleName = classStyleMap.get(agentName);

			Class agentClass = Class.forName(agentName, true, this.getClass().getClassLoader());
			Class styleClass = Class.forName(styleName, true, this.getClass().getClassLoader());

			StyleGIS3D style = (StyleGIS3D) styleClass.newInstance();
			display.registerStyle(agentClass, style);
		}
	}
	
	/**
	 * Method to allow network 2D styles
	 * @param display the 2D display
	 * @param context the simulation context
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 */
	private void registerNetworkStyles(Display2D display, Context context) throws 
	ClassNotFoundException, IllegalAccessException,	InstantiationException {
		
	  String netStyleName = "anl.msc.displays.ScenarioNetStyle2D";
		
	  Network network = (Network)context.getProjection("Steps");

	  Class styleClass = Class.forName(netStyleName, true, this.getClass().getClassLoader());

	  //here the edge style is made and registered with the network
	  EdgeStyle2D style = (EdgeStyle2D) styleClass.newInstance();

	  display.registerNetworkStyle(network, style);
	}
	
	/**
	 * Method to create the visual layout
	 * @param context the simulation context
	 * @return a layout object
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 */
	private Layout createLayout(Context context) throws ClassNotFoundException, 
	  IllegalAccessException, InstantiationException {
    
		String layoutClassName = "repast.simphony.visualization.visualization2D.layout.TreeLayout2D";
    
		//get the class loader
		Class clazz = Class.forName(layoutClassName, true, this.getClass().getClassLoader());
   
		if (!Layout.class.isAssignableFrom(clazz)) {
			// @TODO better errors.
			throw new IllegalArgumentException(
              	"Layout class must implements Layout interface.");
		}
    
		//get the layout
		Layout layout = (Layout) clazz.newInstance();
    
		String layoutProj = "Steps";
    
		Projection<?> projection = context.getProjection(layoutProj);
		if (projection == null) {
			throw new RuntimeException("Projection '" + layoutProj + "' not found.");
		}
		layout.setProjection(projection);

		return layout;
  }

	/**
	 * Properties class for the grid.
	 * 
	 * @author Eric Tatara
	 *
	 */
	class GridProperties implements VisualizationProperties {

		private Map<String,Object> map = new HashMap<String,Object>();
		
		public Iterable<String> getKeys() {
			return map.keySet();
		}

		public Object getProperty(String key) {
			return map.get(key);
		}
		
		public void addProperty(String key, Object value){
			map.put(key, value);
		}
	}
}