package ucl.salinization.model.physical;

import java.util.HashMap;
import java.util.Map;

import cern.jet.random.Normal;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.random.RandomHelper;
import ucl.salinization.model.physical.HydrologicalUnit;
import ucl.salinization.util.Statistics;

/**
 * Model for salt action affecting hydrology/fields.
 * @author Mark Altaweel
 */
public class SalinizationModel {
	
	private double irrigationSalinity = 0;  //amount of irrigation salinity
	
	double rainfall=0;  //rainfall from rain and runoff
	double panEvaporation=0;  //the pan coefficient value
	double stdPanCoefficient=0d; //std for pan coefficient
	double yieldResponseFactor=0; //yield response factor
	double empiricalCoefficient=0d; //empirical coefficient
	VolumeEstimatorModel vem = new VolumeEstimatorModel(); //markov chain model to estimate rainfall
	Map<Crop,Double> cropCoeff=new HashMap<Crop,Double>();  //crop coefficient container
	public static Map<String,Map<Integer,double[]>>seasonalRainfall=new HashMap<String,Map<Integer,double[]>>();
	Map<HydrologicalUnit,Double>concentrationTime=new HashMap<HydrologicalUnit,Double>();
	Map<HydrologicalUnit,Double>initialSalt=new HashMap<HydrologicalUnit,Double>();
	Map<HydrologicalUnit,Double>waterSoil=new HashMap<HydrologicalUnit,Double>();
	Map<HydrologicalUnit,Double>percSoil=new HashMap<HydrologicalUnit,Double>();
	Map<HydrologicalUnit,Double>leachingEff=new HashMap<HydrologicalUnit,Double>();
	Map<String,Integer>rainfallYear=new HashMap<String,Integer>();
	Map<Integer,Integer>panYear=new HashMap<Integer,Integer>();
	private double rainfallSalinity;  //salinity from rainfall
	private Normal normalPan; //pan evap. distribution
	private Normal yieldResponse; //response of yield normal distribution
	private double saltTolerance; //tolerance of salt
	private double fallowScaling;  //scale of fallowing
	static double salinityBasin=0.0;
	static double basinLeaching=0.0;
	static double basinEfficiency=0;
	static double basinStandardDeviation=0;
	static double basinTableSalt=0;
	static double basinUptake=0;
	
	int months[] = {0,1,2,3,4,5,6,7,8,9,10,11};  //months in a year
	
	/**
	 * Main constructor to starting model.
	 */
	public SalinizationModel(){
		double yieldR=(Double)RunEnvironment.getInstance().getParameters().getValue("yieldResponseFactor");
		double yieldRSD=(Double)RunEnvironment.getInstance().getParameters().getValue("yieldResponseFactorSD");
		yieldResponse=RandomHelper.createNormal(yieldR, yieldRSD);
	}

	/**
	 * Method for determining rainfall.
	 */
	public void rainfallRunoff(Field f){
/*		Calendar cal = new GregorianCalendar();
		double tick=
			RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		double daysPerTick=(Double)
			RunEnvironment.getInstance().getParameters().getValue("daysPerTick");
		double day=daysPerTick*tick;
		if(day>365.25)
			day=day%365.25;
		cal.set(Calendar.DAY_OF_YEAR, (int)day); 
		int month=cal.get(Calendar.MONTH); */
		
		String name=f.getName();
		double yearlyRain=0d;
		for(int month:months){
			double[] rainfall=seasonalRainfall.get(name).get(month);
			yearlyRain+=
				Math.max(0, Statistics.computeMean(vem.estimateDischargeValue(rainfall),0,rainfall.length));
			
		}
//		System.out.println(name +" "+yearlyRain+" "+checkYear());
		this.rainfall=yearlyRain/1000;
	}
	
	/**
	 * Method for determining water used for crops.
	 * @param f
	 * @return
	 */
	public double cropWater(Field f){
		double cropW=0;
		
		double cCoeff=RandomHelper.createNormal(f.getCropCoeff(),
				f.getStdCropCoeff()).nextDouble();
		
		if(!panYear.containsKey(checkYear())){
			this.panEvaporation=normalPan.nextDouble();
			panYear.put(checkYear(),checkYear());
		}
		 
		cropW=0.85*cCoeff*this.panEvaporation;
		
		return cropW;
	}
	
	/**
	 * Method for determining irrigation water salinity.
	 * @param hu
	 * @param f
	 * @return
	 */
	public double appliedWater(HydrologicalUnit hu,Field f){
		double appliedIrrigation=0;
	
		this.yieldResponseFactor=this.yieldResponse.nextDouble();
		
		double cWater=cropWater(f);
		appliedIrrigation=cWater/this.yieldResponseFactor*(1-f.getLeachingFactor())
			*Math.max((f.getYield()+this.yieldResponseFactor-1),0);
		
		return appliedIrrigation;
	}

	/**
	 * Readjust leaching factor and other values based on content of salinization in basin fields
	 * @param f the field to adjust
	 * @param earlyS
	 */
	public void calculateLeachingFactor(Field f){
		double lf = f.getLeachingFactor();
		double newPRZ=f.getAppliedWaterSalinity()/lf;
		f.setSoilWaterSalinity(newPRZ);
		double soilWaterSalinity=f.getSoilWaterSalinity();
		double req=f.getAppliedWaterSalinity()/soilWaterSalinity;
		f.setLeachingFactor(req);
		if(salinityBasin>60 && basinLeaching<lf && f.getFieldType()!=1){
		//	if(f.getLeachingFactor()*(earlyS/f.getSaltCompounds())>0)
				f.setLeachingFactor((basinLeaching+f.getLeachingFactor())/2);
	//		    System.out.println(f.getId() + ":" + f.getLeachingFactor() + " : "+f.getFieldType());
			    f.setLeachingEfficiency((basinEfficiency+f.getLeachingEfficiency())/2.0);
			    f.setLeachingSTD((basinStandardDeviation+f.getLeachingSTD())/2.0);
			    f.setWaterTableSalt((basinTableSalt+f.getWaterTableSalt())/2.0);
			    f.setRatioUptake((basinUptake+f.getRatioUptake())/2.0);
		}
		
	}
	/**
	 * Method for determining the percent of salt removed from fields.
	 * 
	 *Conceptually, the method does the following but slightly more simplified:
     *
     * Ct = Ci + (Co - Ci) exp (-EL.T.Qf/Ws)
     *
	 *where C = salt concentration, 
	 *Ct = C in the soil at time T, 
	 *Co = C in the soil at time T=0, 
	 *Ci = C of the irrigation water, 
	 *EL = leaching efficiency, 
	 *Qp = average percolation rate through the soil, 
	 *and Ws = water stored in the soil at field saturation.
	 *
	 * @param field a field object.
	 */
	public void drainSalt(Field field){
//		double saltRemoval=(Double)RunEnvironment.getInstance().getParameters().getValue("saltRemoval");
//		double std=(Double)RunEnvironment.getInstance().
//			getParameters().getValue("saltRemovalSD");
		double s1=field.getSaltCompounds();
		double saltRemoval=field.getLeachingFactor();
		double std=field.getLeachingSTD();
		double sRemove=Math.max(0,field.getLeachingEfficiency()*RandomHelper.createNormal(saltRemoval,std).nextDouble());
/*		
	//	sRemove=Math.min(sRemove, 1.0);
	//	if(sRemove>1.0 || 0>sRemove){
	//		System.out.println(sRemove);
	//	}
	//	double v=Math.log(1+sRemove);
		double v=Math.max(0,(Math.log(Math.E+(field.getSaltCompounds())))/(Math.log(sRemove*(Math.E+(field.getSaltCompounds())))));
		field.setSaltCompounds(field.getSaltCompounds()*v); 
	//	field.setSaltCompounds(field.getSaltCompounds()*(1-sRemove)); */
	//	field.setSaltCompounds(Math.max(0,field.getSaltCompounds()*v)); 
//		double s = field.getSaltCompounds();
		
//		if(field.getId()==101){
//			System.out.println("stop");
//		}
		
//		if(field.getStatus()==Field.FALLOW_FIELD){
//			field.setAppliedWaterSalinity(0.0d);
//		}
//		double salt=field.getAppliedWaterSalinity()+(field.getInitialSalinization() - field.getAppliedWaterSalinity())
//			*Math.exp(-field.getLeachingEfficiency()*checkYear()*field.getPercolationRate()/field.getWaterStoredSaturation());
//		double salt=field.getAppliedWaterSalinity()+(field.getSaltCompounds() - field.getAppliedWaterSalinity())
//			*Math.exp(-field.getLeachingEfficiency()*1*field.getPercolationRate()/field.getWaterStoredSaturation());
		
//		field.setSaltCompounds(salt);
		double f=0;
		double s=0;
		double sR=sRemove;
		if(field.getFallowYears()>0){
			f=field.getTrueFallowYears()-field.getFallowYears();
			s=field.getInitialSalinization();
	//		if(f>0)
	//			sR=field.getsRemove();
		}
		else{
			f=field.getFallowYears();
			s=field.getSaltCompounds();
		}
		double salt=s*Math.exp(-sR*(f+1));
//		if(salt<0){
//			System.out.println("stop");
//		}	
		field.setSaltCompounds(salt);
		double ns=s1/salt;
		
		//modification to leaching and other factors based on salinity content in fields
		if(salinityBasin<60 && field.getFieldType()!=1){
			field.setLeachingFactor((field.getLeachingFactor()+field.getInitialLeachingFactor())/2.0);
			field.setLeachingSTD((field.getInitialStandardDeviation()+field.getLeachingSTD())/2.0);
			field.setLeachingEfficiency((field.getInitialEfficiency()+field.getLeachingEfficiency())/2.0);
			field.setWaterTableSalt((field.getInitialWaterTable()+field.getWaterTableSalt())/2.0);
		    field.setRatioUptake((field.getInitialRatioUptake()+field.getRatioUptake())/2.0);
//			System.out.println(field.getLeachingFactor());
		}
//		if(salinityBasin>50 && field.getLeachingFactor()>salinityBasin){
//			field.setLeachingFactor((field.getLeachingFactor()+basinLeaching)/2);
//		}
	//	field.setsRemove(sRemove);
	}
	
	/**
	 * Method to check what year is it.
	 * @return the year value.
	 */
	public int checkYear(){
		double tick=RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		double daysPT=(Double)RunEnvironment.getInstance().getParameters().getValue("daysPerTick");
	
		int year=(int) Math.round(tick*daysPT/365.25);
		
		return year;
	}
	
	/**
	 * Method for determining irrigation water applied.
	 * @param f
	 * @return
	 */
	public double irrigationWater(Field f){
		double irrigW=0;
		HydrologicalUnit hu=f.getFieldWater();
		irrigW=Math.max(appliedWater(hu,f)-this.rainfall,0);
		
		return irrigW;
	}
	
	/**
	 * Method for determining root salinity.
	 * @param f
	 * @return rootSalinity the root salinity
	 */
	public double rootSalinity(Field f){
		double rootSalinity=0d;
		double appWS=appliedWaterSalinity(f);
		rootSalinity=(0.5*this.empiricalCoefficient*appWS
			*(1+1/f.getLeachingFactor()));
	
		f.setSaltCompounds(Math.max(0,f.getSaltCompounds()+rootSalinity)); 
		return f.getSaltCompounds();
	}
	
	/**
	 * This is the capillary rise function. Which is stated as:
	 * 
	 * ECe,i=ECe,i-1+(1.5J(ECwt)Si,l)/d
	 * 
	 * 
	 * @param f the field input
	 */
	public void capilaryRise(Field f){
		double capSalt=RandomHelper.createNormal(f.getCapillarySalt(), f.getCapillarySaltSD()).nextDouble();
		
//		f.setCapillarySalt(capSalt);
//		f.setSaltCompounds(f.getSaltCompounds()+capSalt);
		
		double wt=f.getWaterTableSalt();
		double ratioUptake=f.getRatioUptake();
		double d=f.getLayerThickness();// i.e., 1
		
		double ecC=(f.getSaltCompounds()/2.0)+((1.5*capSalt*wt*ratioUptake)/d);
		if(ecC<0){
			System.out.println("stop");
		}
		f.setSaltCompounds(ecC*2.0);
	}
	
	/**
	 * Method for determining how saline water used for farming is.
	 * @param f the agricultural field.
	 * @return a saline value for water.
	 */
	public double appliedWaterSalinity(Field f){
		double salinity=0d;
		
		double irrigationDepth=irrigationWater(f);
		
	
		salinity=(this.rainfall*this.rainfallSalinity+irrigationDepth*this.irrigationSalinity)/
			(this.rainfall+irrigationDepth);
		
		f.setAppliedWaterSalinity(salinity);
		return salinity;
	}
	
	/**
	 * Crop yield response relative to root zone salinity and leaching.
	 * @param f the field
	 * @param rootSalinity the root zone salinity.
	 */
	public void cropYield(Field f,double rootSalinity){
		double yieldPercentage=0;
		double threshSalinity=(Double) RunEnvironment.getInstance().getParameters().getValue("thresholdSalinity");
		double yieldReduction=(Double)RunEnvironment.getInstance().getParameters().getValue("yieldReduction");
		
		yieldPercentage=100-yieldReduction*((rootSalinity/2.0)-threshSalinity);
		
		f.setYield(Math.min(yieldPercentage/100,1.0));
		f.setYield(Math.max(0,f.getYield()));
	}
	
	/**
	 * Method to run model.
	 * @param f the field used to calculate salinity.
	 */
	public void run(Field f){
		double earlyS=f.getSaltCompounds();
		String y=Integer.toString(checkYear())+f.getName();
		if(!rainfallYear.containsKey(y)){
			rainfallRunoff(f);
			rainfallYear.put(y, checkYear());
		}
		double rootSalinity=rootSalinity(f);
		cropYield(f,rootSalinity);
		calculateLeachingFactor(f);  //update leaching factor for next cycle
		if(f.getFieldType()==1){
			if(f.getSaltCompounds()>salinityBasin){
				salinityBasin=f.getSaltCompounds();
				basinLeaching=f.getLeachingFactor();
				basinEfficiency=f.getLeachingEfficiency();
				basinStandardDeviation=f.getLeachingSTD();
				basinUptake=f.getRatioUptake();
				basinTableSalt=f.getWaterTableSalt();
			}
		}
//		System.out.println(basinLeaching);
	}
	
	//some getters and setters below
	
	public double getPanCoefficient() {
		return panEvaporation;
	}

	public void setPanCoefficient(double panCoefficient) {
		this.panEvaporation = panCoefficient;
	}

	public double getEmpiricalCoefficient() {
		return empiricalCoefficient;
	}

	public void setEmpiricalCoefficient(double empiricalCoefficient) {
		this.empiricalCoefficient = empiricalCoefficient;
	}
	
	public double getRainfallSalinity() {
		return rainfallSalinity;
	}

	public void setRainfallSalinity(double rainfallSalinity) {
		this.rainfallSalinity = rainfallSalinity;
	}

	
	public double getIrrigationSalinity() {
		return irrigationSalinity;
	}

	public void setIrrigationSalinity(double irrigationSalinity) {
		this.irrigationSalinity = irrigationSalinity;
	}

	public void setNormalPanEvaporation(Normal normalPan){
		this.normalPan=normalPan;
	}
	
	public double getSaltTolerance() {
		return saltTolerance;
	}

	public void setSaltTolerance(double saltTolerance) {
		this.saltTolerance = saltTolerance;
	}

	public double getFallowScaling() {
		return fallowScaling;
	}

	public void setFallowScaling(double fallowScaling) {
		this.fallowScaling = fallowScaling;
	}

	public static void main(String[] args){
		
	}
}
