package ucl.salinization.util;

/**
 * A number of static methods to compute statistical properties of an
 * array of double values.
 * Implements the computation of mean, variance and standard deviation
 * for <code>double</code> values.
 * @author (added a method)
 * @since 0.11.0
 */
public class Statistics
{
	private Statistics()
	{
	}

	/**
	 * Computes the mean value for the argument array.
	 * Adds all values and divides them by the number of array elements.
	 * @param values double array on which the mean is to be determined
	 * @return computed mean value
	 * @throws IllegalArgumentException if the array has not at least one element
	 */
	public static double computeMean(double[] values)
	{
		return computeMean(values, 0, values.length);
	}

	/**
	 * Computes the mean value for some elements of the argument array.
	 * Adds all values and divides them by the number of array elements.
	 * @param values array from which elements are read
	 * @param offset index of the first element to be used
	 * @param number number of elements to be used
	 * @return computed mean value
	 * @throws IllegalArgumentException if the array has not at least one element
	 */
	public static double computeMean(double[] values, int offset, int number)
	{
		if (number < 1)
		{
			throw new IllegalArgumentException("The number of values to process must be one or larger.");
		}
		double sum = 0;
		final int UNTIL = offset + number;
		do
		{
			sum += values[offset++];
		}
		while (offset != UNTIL);
		return sum / number; 
	}
	
	/**
	 * Method to compute the sum of a double array.
	 * @param values the double array
	 * @return a sum of the array values
	 */
	public static double computeSum(double[] values)
	{
		double sum=0;
		for(int i=0; i < values.length;i++){
			sum+=values[i];
		}
		return sum;
	}
	
	/**
	 * Method to compute the sum of a double array.
	 * @param values the double array
	 * @return a sum of the array values
	 */
	public static double computeSum(int[] values)
	{
		double sum=0;
		for(int i=0; i < values.length;i++){
			sum+=values[i];
		}
		return sum;
	}

	/**
	 * Computes the standard deviation for the argument array of values.
	 * @param values array from which elements are read
	 * @return computed standard deviation
	 * @throws IllegalArgumentException if the array has not at least two elements
	 */
	public static double computeStandardDeviation(double[] values)
	{
		return computeStandardDeviation(values, 0, values.length);
	}

	/**
	 * Computes the standard deviation for the argument array of values.
	 * Reuses the mean value for that argument which must have been computed before.
	 * @param values array from which elements are read
	 * @param mean the mean value for the array, possibly computed with a 
	 *   call to {@link #computeMean(double[])}.
	 * @return computed standard deviation
	 * @throws IllegalArgumentException if the array has not at least two elements
	 */
	public static double computeStandardDeviation(double[] values, double mean)
	{
		return computeStandardDeviation(values, 0, values.length, mean);
	}

	/**
	 * Computes the standard deviation for some of the argument array's values.
	 * If you already have computed a mean value using {@link #computeMean(double[], int, int)},
	 * better call {@link #computeStandardDeviation(double[], int, int, double)}.
	 * Otherwise, this method has to compute mean again.
	 * @param values array from which elements are read
	 * @param offset first element to be used
	 * @param number number of elements used starting at values[offset]
	 * @return computed standard deviation
	 * @throws IllegalArgumentException if the array has not at least two elements
	 */
	public static double computeStandardDeviation(double[] values, int offset, int number)
	{
		double mean = computeMean(values, offset, number);
		return computeStandardDeviation(values, 0, values.length, mean);
	}

	/**
	 * Computes the standard deviation for some of the argument array's values.
	 * Use this version of the method if you already have a mean value,
	 * otherwise this method must be computed again.
	 * @param values array from which elements are read
	 * @param offset first element to be used
	 * @param number number of elements used starting at values[offset]
	 * @param mean value of the elements
	 * @return computed standard deviation
	 * @throws IllegalArgumentException if the array has not at least two elements
	 */
	public static double computeStandardDeviation(double[] values, int offset, int number, double mean)
	{
		return Math.sqrt(computeVariance(values, offset, number, mean));
	}

	/**
	 * Computes the variance for the argument array.
	 * @param values array from which elements are read
	 * @return variance for the array elements
	 * @throws IllegalArgumentException if the array has not at least two elements
	 */
	public static double computeVariance(final double[] values)
	{
		return computeVariance(values, 0, values.length);
	}

	/**
	 * Computes the variance for some of the argument array's values.
	 * @param values array from which elements are read
	 * @param mean the mean for the array elements
	 * @return variance for the array elements
	 * @throws IllegalArgumentException if the array has not at least two elements
	 */
	public static double computeVariance(final double[] values, final double mean)
	{
		return computeVariance(values, 0, values.length, mean);
	}

	/**
	 * Computes the variance for some of the argument array's values.
	 * If you already have computed a mean value using {@link #computeMean(double[], int, int)},
	 * better call {@link #computeVariance(double[], int, int, double)}.
	 * Otherwise, this method has to compute mean again.
	 * @param values array from which elements are read
	 * @param offset first element to be used
	 * @param number number of elements used starting at values[offset]
	 * @return computed variance
	 * @throws IllegalArgumentException if the array has not at least two elements
	 */
	public static double computeVariance(final double[] values, int offset, final int number)
	{
		double mean = computeMean(values, offset, number);
		return computeVariance(values, 0, values.length, mean);
	}

	/**
	 * Computes the variance for some of the argument array's values.
	 * Use this version of the method in case mean has already been
	 * computed.
	 * @param values array from which elements are read
	 * @param offset first element to be used
	 * @param number number of elements used starting at values[offset]
	 * @param mean the mean for the array elements
	 * @return computed variance
	 * @throws IllegalArgumentException if the array has not at least two elements
	 */
	public static double computeVariance(final double[] values, int offset, final int number, final double mean)
	{
		if (number < 2)
		{
			throw new IllegalArgumentException("The number of values to process must be two or larger.");
		}
		double sum = 0;
		final int UNTIL = offset + number;
		do
		{
			double diff = values[offset++] - mean;
			sum += diff * diff;
		}
		while (offset != UNTIL);
		return sum / (number - 1); 
	}
	
	public static double[] combineData(double values1[],double values2[]){
		double returnValues[]=new double[values1.length+values2.length];
		for(int i=0; i<values1.length;i++){
			returnValues[i]=values1[i];
		}
		for(int j=0; j<values2.length;j++){
			returnValues[j+(values1.length-1)]=values2[j];
		}
		return returnValues;
	}
	
	/**
	 * Exchange two elements in the argument array.
	 * A temporary variable is used so that a[i1] will
	 * hold the value that was previously stored at a[i2]
	 * and vice versa.
	 * 
	 * @param a the array in which two elements are swapped
	 * @param i1 index of the first element
	 * @param i2 index of the second element
	 * @throws ArrayIndexOutOfBoundsException if either i1 or i2 are
	 *  not valid index values into a (from 0 to a.length - 1)
	 */
	public static void swap(double[] a, int i1, int i2)
	{
		double temp = a[i1];
		a[i1] = a[i2];
		a[i2] = temp;
	}
	
	/**
	 * Find the median value of the specified interval of the argument array.
	 * The interval starts at index <code>from</code> and goes to
	 * <code>to</code>; the values at these positions are included.
	 * Note that the array will be modified while searching, so you might want
	 * to backup your data.
	 * <p>
	 * This implementation is a port of the C function from
	 * <code>quickselect.c</code>, provided at <a target="_top" 
	 * href="http://ndevilla.free.fr/median/">http://ndevilla.free.fr/median/</a>.
	 * The page is a good resource for various median value algorithms, including
	 * implementations and benchmarks.
	 * <p>
	 * The original code on which this class is based was written in C++
	 * by Martin Leese.
	 * It was ported to C and optimized by Nicolas Devillard (author of the 
	 * above mentioned page).
	 * The algorithm is from <em>Numerical recipes in C</em>, Second Edition,
     * Cambridge University Press, 1992, Section 8.5, ISBN 0-521-43108-5.
	 * <p>
	 *
	 * @param a the array
	 * @param from the index of the start of the interval in which the median value will be searched
	 * @param to the index of the end of the interval in which the median value will be searched
	 * @return the median value
	 */
	public static double computeMedian(double[] a, int from, int to)
	{
		int low = from;
		int high = to;
		int median = (low + high) / 2;
		do
		{
			if (high <= low)
			{
				return a[median];
			}
			if (high == low + 1)
			{
				if (a[low] > a[high])
				{
					swap(a, low, high);
				}
				return a[median];
			}
			int middle = (low + high) / 2;
			if (a[middle] > a[high])
			{
				swap(a, middle, high);
			}
			if (a[low] > a[high])
			{
				swap(a, low, high);
			}
			if (a[middle] > a[low])
			{
				swap(a, middle, low);
			}
			swap(a, middle, low + 1);
			int ll = low + 1;
			int hh = high;
			do
			{
				do
				{
					ll++;
				}
				while(a[low] > a[ll]);
				do
				{
					hh--;
				}
				while(a[hh] > a[low]);
				if (hh < ll)
				{
					break;
				}
				swap(a, ll, hh);
			}
			while(true);
			swap(a, low, hh);
			if (hh <= median)
			{
				low = ll;
			}
			if (hh >= median)
			{
				high = hh - 1;
			}
		}
		while(true);
	}
}
